@@ -1,17 +1,19 @@
 """Bucket implementation using SQLite"""
+
+import logging
 import sqlite3
 from contextlib import nullcontext
 from pathlib import Path
 from tempfile import gettempdir
 from threading import RLock
-from typing import List
-from typing import Optional
-from typing import Tuple
-from typing import Union
+from time import time, time_ns
+from typing import List, Optional, Tuple, Union
 
-from ..abstracts import AbstractBucket
-from ..abstracts import Rate
-from ..abstracts import RateItem
+from ..abstracts import AbstractBucket, Rate, RateItem
+from ..clocks import AbstractClock
+from ..utils import dedicated_sqlite_clock_connection
+
+logger = logging.getLogger(__name__)
 
 
 class Queries:
@@ -42,9 +44,7 @@ class Queries:
     DROP_INDEX = "DROP INDEX IF EXISTS '{index}'"
     COUNT_ALL = "SELECT COUNT(*) FROM '{table}'"
     GET_ALL_ITEM = "SELECT * FROM '{table}' ORDER BY item_timestamp ASC"
-    GET_FIRST_ITEM = (
-        "SELECT name, item_timestamp FROM '{table}' ORDER BY item_timestamp ASC"
-    )
+    GET_FIRST_ITEM = "SELECT name, item_timestamp FROM '{table}' ORDER BY item_timestamp ASC"
     GET_LAG = """
     SELECT (strftime ('%s', 'now') || substr(strftime ('%f', 'now'), 4)) - (
     SELECT item_timestamp
@@ -68,19 +68,30 @@ class SQLiteBucket(AbstractBucket):
     table: str
     full_count_query: str
     lock: RLock
+    use_limiter_lock: bool
 
-    def __init__(
-        self, rates: List[Rate], conn: sqlite3.Connection, table: str, lock=None
-    ):
+    def __init__(self, rates: List[Rate], conn: sqlite3.Connection, table: str, lock=None):
         self.conn = conn
         self.table = table
         self.rates = rates
 
         if not lock:
+            self.use_limiter_lock = False
             self.lock = RLock()
         else:
+            self.use_limiter_lock = True
             self.lock = lock
 
+    def now(self):
+        # TODO: Use Sqlite time source via a Lua script
+        return time_ns() // 1000000
+
+    def limiter_lock(self):
+        if self.use_limiter_lock:
+            return self.lock
+        else:
+            return None
+
     def _build_full_count_query(self, current_timestamp: int) -> Tuple[str, dict]:
         full_query: List[str] = []
 
@@ -91,9 +102,7 @@ class SQLiteBucket(AbstractBucket):
             query = Queries.COUNT_BEFORE_INSERT.format(table=self.table, index=index)
             full_query.append(query)
 
-        join_full_query = (
-            " union ".join(full_query) if len(full_query) > 1 else full_query[0]
-        )
+        join_full_query = " union ".join(full_query) if len(full_query) > 1 else full_query[0]
         return join_full_query, parameters
 
     def put(self, item: RateItem) -> bool:
@@ -113,9 +122,7 @@ class SQLiteBucket(AbstractBucket):
                     self.failing_rate = rate
                     return False
 
-            items = ", ".join(
-                [f"('{name}', {item.timestamp})" for name in [item.name] * item.weight]
-            )
+            items = ", ".join([f"('{name}', {item.timestamp})" for name in [item.name] * item.weight])
             query = (Queries.PUT_ITEM.format(table=self.table)) % items
             self.conn.execute(query).close()
             self.conn.commit()
@@ -146,9 +153,7 @@ class SQLiteBucket(AbstractBucket):
 
     def count(self) -> int:
         with self.lock:
-            cur = self.conn.execute(
-                Queries.COUNT_ALL.format(table=self.table)
-            )
+            cur = self.conn.execute(Queries.COUNT_ALL.format(table=self.table))
             ret = cur.fetchone()[0]
             cur.close()
             return ret
@@ -165,21 +170,26 @@ class SQLiteBucket(AbstractBucket):
 
             return RateItem(item[0], item[1])
 
+    def close(self):
+        with self.lock:
+            if self.conn is not None:
+                try:
+                    self.conn.close()
+                    self.conn = None
+                except Exception as e:
+                    logger.debug("Exception %s closing sql connection", e)
+
     @classmethod
     def init_from_file(
-        cls,
-        rates: List[Rate],
-        table: str = "rate_bucket",
-        db_path: Optional[str] = None,
-        create_new_table: bool = True,
-        use_file_lock: bool = False
+        cls, rates: List[Rate], table: str = "rate_bucket", db_path: Optional[str] = None, create_new_table: bool = True, use_file_lock: bool = False
     ) -> "SQLiteBucket":
         if db_path is None and use_file_lock:
             raise ValueError("db_path must be specified when using use_file_lock")
 
         if db_path is None:
             temp_dir = Path(gettempdir())
-            db_path = str(temp_dir / "pyrate_limiter.sqlite")
+
+            db_path = str(temp_dir / f"pyrate_limiter_{time()}.sqlite")
 
         # TBD: FileLock switched to a thread-local FileLock in 3.11.0.
         # Should we set FileLock's thread_local to False, for cases where user is both multiprocessing & threading?
@@ -191,19 +201,14 @@ class SQLiteBucket(AbstractBucket):
         if use_file_lock:
             try:
                 from filelock import FileLock  # type: ignore[import-untyped]
+
                 file_lock = FileLock(db_path + ".lock")  # type: ignore[no-redef]
                 file_lock_ctx: Union[nullcontext, FileLock] = file_lock  # type: ignore[no-redef]
-            except ImportError:
-                raise ImportError(
-                    "filelock is required for file locking. "
-                    "Please install it as optional dependency"
-                )
+            except ImportError as e:
+                raise ImportError("filelock is required for file locking. Please install it as optional dependency") from e
 
         with file_lock_ctx:
             assert db_path is not None
-            assert db_path.endswith(".sqlite"), (
-                "Please provide a valid sqlite file path"
-            )
 
             sqlite_connection = sqlite3.connect(
                 db_path,
@@ -220,12 +225,10 @@ class SQLiteBucket(AbstractBucket):
                 cur.execute("PRAGMA synchronous=NORMAL;")
 
             if create_new_table:
-                cur.execute(
-                    Queries.CREATE_BUCKET_TABLE.format(table=table)
-                )
+                cur.execute(Queries.CREATE_BUCKET_TABLE.format(table=table))
 
             create_idx_query = Queries.CREATE_INDEX_ON_TIMESTAMP.format(
-                index_name="idx_rate_item_timestamp",
+                index_name=f"idx_{table}_rate_item_timestamp",
                 table_name=table,
             )
 
@@ -234,3 +237,34 @@ class SQLiteBucket(AbstractBucket):
             sqlite_connection.commit()
 
             return cls(rates, sqlite_connection, table=table, lock=file_lock)
+
+
+class SQLiteClock(AbstractClock):
+    """Get timestamp using SQLite as remote clock backend"""
+
+    time_query = "SELECT CAST(ROUND((julianday('now') - 2440587.5)*86400000) As INTEGER)"
+
+    def __init__(self, conn: Union[sqlite3.Connection, SQLiteBucket]):
+        """
+        In multiprocessing cases, use the bucket, so that a shared lock is used.
+        """
+
+        self.lock: Optional[RLock] = None
+
+        if isinstance(conn, SQLiteBucket):
+            self.conn = conn.conn
+            self.lock = conn.lock
+        else:
+            self.conn = conn
+
+    @classmethod
+    def default(cls):
+        conn = dedicated_sqlite_clock_connection()
+        return cls(conn)
+
+    def now(self) -> int:
+        with self.lock if self.lock else nullcontext():
+            cur = self.conn.execute(self.time_query)
+            now = cur.fetchone()[0]
+            cur.close()
+            return int(now)
